
(define-extension icalendar)

(require-extension syntax-case)
(require-extension matchable)
(require-extension srfi-1)
(require-extension srfi-13)
(require-extension srfi-14)

(declare (export icalendar:parse))


(define-record token symbol value line)

(define (token p line)
  (cons (car p)
	(match p
	       [(sym val) (make-token sym val line)]
	       [(sym) (make-token sym #f line)])))

(define-syntax tok
  (syntax-rules ()
    ((tok t) (token (quasiquote t) 0))))

(define-record-printer (token x out)
  (fprintf out "#(token ~A value=~S)"
	   (token-symbol x) 
	   (token-value x) ))


(define (parse-error msg #!optional arg)
  (match arg
	 [#f (error msg)]
	 [(or (_ . ($ token symbol value line))
	      ($ token symbol value line))
	  (error (conc "line " line ": " msg)
		 (conc symbol (if value (conc " " value) "")))]
	 [_ (error msg arg)]))

(define not-safe-charset 
  (char-set-union char-set:iso-control (char-set #\" #\; #\: #\,)))

(define (safe-char? c)
  (and (char? c) (not (char-set-contains? not-safe-charset c))))

(define (char-alphanumeric? c)
  (and (char? c) (or (char-alphabetic? c) (char-numeric? c))))

(define (char-blank? c)
  (and (char? c) (char-set-contains? char-set:blank c)))

(define (value-char? c)
  (and (char? c) (char-set-contains? char-set:printing c)))

(define lexer-mode #f)

(define (make-lexer port errorp)
  (lambda ()
    (letrec (
	     (line-unfold
	      (lambda (c v k)
		(if (and (char? c) (char=? c #\return) )
		    (begin
		      (read-char port)
		      (if (char=? #\linefeed (peek-char port))
			  (let ((lf (read-char port))
				(w  (peek-char port)))
			    (if (char-blank? w)
				(let* ((w (read-char port))
				       (n (peek-char port)))
				  (k n))
				(v w)))
			  (begin (errorp "line-unfold: illegal character" c)
				 (k (read-char port)))))
		    (k c))))

	     (read-qstring
	      (lambda (cs)
		(let ([c (read-char port)])
		  (cond [(eq? 'eof c)    (errorp "unexpected end of string constant") ]
			[(char=? c #\return)  
			 (if (char=? #\linefeed (peek-char port))
			     (begin (read-char port)
				    (if (char=? #\space (peek-char port))
					(begin (read-char port)
					       (read-qstring cs))
					(read-qstring (cons* #\linefeed #\return cs))))
			     (read-qstring (cons #\return cs)))]
			[(char=? c #\\)  (read-qstring (cons (read-char port) cs)) ]
			[(char=? c #\")  (reverse-list->string cs) ]
			[else (read-qstring (cons c cs))]))))

	     (read-sstring
	      (lambda (cs)
		(let ([c (peek-char port)])
		  (line-unfold c (lambda (w) (begin (set! lexer-mode #f) (reverse-list->string cs)))
			 (lambda (c)
			   (cond [(safe-char? c)   (read-sstring (cons (read-char port) cs)) ]
				 [else             (begin (set! lexer-mode #f)
							  (reverse-list->string cs)) ]))
			 ))))

	     (read-vstring
	      (lambda (cs)
		(let ([c (peek-char port)])
		  (line-unfold c (lambda (w) (begin (set! lexer-mode #f) (reverse-list->string cs)))
			 (lambda (c)
			     (cond [(value-char? c)  (read-vstring (cons (read-char port) cs)) ]
				   [else             (begin (set! lexer-mode #f)
							    (reverse-list->string cs)) ]))
			 ))))

             (read-id
              (lambda (l)
                (let ((c (peek-char port)))
                  (if (and (char? c) (or (char-alphanumeric? c) (char=? c #\-)))
                      (read-id (cons (read-char port) l))
                      (string->symbol (string-upcase (apply string (reverse l)))))))))

      ;; -- read the next token
      (let loop ((c (peek-char port)))
	(let ((t (line-unfold c (lambda (w) (begin (set! lexer-mode #f) 'CRLF))
			(lambda (c)
			    (cond
			     ((eof-object? c)       '*eoi*)
			     
			     ((char=? c #\=)        (begin (if (not lexer-mode) (set! lexer-mode 'sstring))
							   (read-char port)
							   'EQ))
			     ((char=? c #\:)        (begin (if (not lexer-mode) (set! lexer-mode 'vstring))
							   (read-char port)
							   'COLON))
			     ((char=? c #\;)        (begin (read-char port) 'SEMICOLON))
			     ((char=? c #\,)        (begin (read-char port) 'COMMA))
			     ((char=? c #\")        (tok (QSTRING ,(read-qstring (list)))))

			     ((and (eq? lexer-mode 'vstring) (value-char? c))
			      (tok (VSTRING ,(read-vstring (list (read-char port))))))

			     ((and (eq? lexer-mode 'sstring) (safe-char? c))
			      (tok (SSTRING ,(read-sstring (list (read-char port))))))

			     ((char-alphabetic? c)  (let ((sym  (read-id (list (read-char port)))))
						      (case sym
							((BEGIN END)  (begin (set! lexer-mode 'container)  sym))
							((VCALENDAR VTODO VEVENT VCARD VJOURNAL VFREEBUSY VALARM ) sym)
							(else (tok (ID ,sym))))))

			     (else
			      (errorp "illegal character" c)
			      (loop (read-char port))))))))
	  t )))))

(include "icalendar.grm.scm")

(define (icalendar:parse s)
  (or (and (string? s) (string-null? s) '())
      (let ((port
	     (cond ((string? s)  (open-input-string s))
		   ((port? s)    s)
		   (else (error 'parse-ical "bad argument type: not a string or a port: " s)))))
	(icalendar-parser  (make-lexer port parse-error) parse-error))))
